@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013-2022  Free Software Foundation, Inc.
+ * Copyright (C) 2013-2023  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -17,6 +17,10 @@
  *	Paulo Cesar Pereira de Andrade
  */
 
+/* callee save
+ * align16(lr+fp+x19+x2[0-8]+v8+v9+v1[0-15]) */
+#define stack_framesize			160
+
 #define jit_arg_reg_p(i)		((i) >= 0 && (i) < 8)
 #define jit_arg_f_reg_p(i)		((i) >= 0 && (i) < 8)
 
@@ -60,6 +64,8 @@ typedef struct jit_va_list {
 /*
  * Prototypes
  */
+#define compute_framesize()		_compute_framesize(_jit)
+static void _compute_framesize(jit_state_t*);
 #define patch(instr, node)		_patch(_jit, instr, node)
 static void _patch(jit_state_t*,jit_word_t,jit_node_t*);
 
@@ -146,6 +152,14 @@ jit_register_t		_rvs[] = {
     { _NOREG,				"<none>" },
 };
 
+static jit_int32_t iregs[] = {
+    _R19, _R20, _R21, _R22, _R23, _R24, _R25, _R26, _R27, _R28
+};
+
+static jit_int32_t fregs[] = {
+    _V8, _V9, _V10, _V11, _V12, _V13, _V14, _V15
+};
+
 /*
  * Implementation
  */
@@ -206,6 +220,7 @@ jit_int32_t
 _jit_allocai(jit_state_t *_jit, jit_int32_t length)
 {
     assert(_jitc->function);
+    jit_check_frame();
     switch (length) {
 	case 0:	case 1:						break;
 	case 2:		_jitc->function->self.aoff &= -2;	break;
@@ -347,6 +362,7 @@ void
 _jit_ellipsis(jit_state_t *_jit)
 {
     jit_inc_synth(ellipsis);
+    jit_check_frame();
     if (_jitc->prepare) {
 	jit_link_prepare();
 	assert(!(_jitc->function->call.call & jit_call_varargs));
@@ -410,6 +426,7 @@ _jit_arg(jit_state_t *_jit, jit_code_t code)
 #else
 	_jitc->function->self.size += sizeof(jit_word_t);
 #endif
+	jit_check_frame();
     }
     node = jit_new_node_ww(code, offset,
 			   ++_jitc->function->self.argn);
@@ -437,6 +454,7 @@ _jit_arg_f(jit_state_t *_jit)
 #else
 	_jitc->function->self.size += sizeof(jit_word_t);
 #endif
+	jit_check_frame();
     }
     node = jit_new_node_ww(jit_code_arg_f, offset,
 			   ++_jitc->function->self.argn);
@@ -460,6 +478,7 @@ _jit_arg_d(jit_state_t *_jit)
 #endif
 	offset = _jitc->function->self.size;
 	_jitc->function->self.size += sizeof(jit_float64_t);
+	jit_check_frame();
     }
     node = jit_new_node_ww(jit_code_arg_d, offset,
 			   ++_jitc->function->self.argn);
@@ -479,8 +498,10 @@ _jit_getarg_c(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	jit_extr_c(u, JIT_RA0 - v->u.w);
 #endif
     }
-    else
-	jit_ldxi_c(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_c(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -496,8 +517,10 @@ _jit_getarg_uc(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	jit_extr_uc(u, JIT_RA0 - v->u.w);
 #endif
     }
-    else
-	jit_ldxi_uc(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_uc(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -513,8 +536,10 @@ _jit_getarg_s(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	jit_extr_s(u, JIT_RA0 - v->u.w);
 #endif
     }
-    else
-	jit_ldxi_s(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_s(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -530,8 +555,10 @@ _jit_getarg_us(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	jit_extr_us(u, JIT_RA0 - v->u.w);
 #endif
     }
-    else
-	jit_ldxi_us(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_us(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -547,8 +574,10 @@ _jit_getarg_i(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	jit_extr_i(u, JIT_RA0 - v->u.w);
 #endif
     }
-    else
-	jit_ldxi_i(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_i(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -565,8 +594,10 @@ _jit_getarg_ui(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	jit_extr_ui(u, JIT_RA0 - v->u.w);
 #endif
     }
-    else
-	jit_ldxi_ui(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_ui(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -577,8 +608,10 @@ _jit_getarg_l(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     jit_inc_synth_wp(getarg_l, u, v);
     if (jit_arg_reg_p(v->u.w))
 	jit_movr(u, JIT_RA0 - v->u.w);
-    else
-	jit_ldxi_l(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_l(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 #endif
@@ -610,26 +643,28 @@ _jit_putargr(jit_state_t *_jit, jit_int32_t u, jit_node_t *v, jit_code_t code)
 #endif
     }
     else {
+	jit_node_t	*node;
 #if PACKED_STACK
 	switch (code) {
 	    case jit_code_putargr_c:	case jit_code_putargr_uc:
-		jit_stxi_c(v->u.w, JIT_FP, u);			break;
+		node = jit_stxi_c(v->u.w, JIT_FP, u);		break;
 	    case jit_code_putargr_s:	case jit_code_putargr_us:
-		jit_stxi_s(v->u.w, JIT_FP, u);			break;
+		node = jit_stxi_s(v->u.w, JIT_FP, u);		break;
 #  if __WORDSIZE == 32
 	    case jit_code_putargr_i:
-		jit_stxi(v->u.w, JIT_FP, u);			break;
+		node = jit_stxi(v->u.w, JIT_FP, u);		break;
 #  else
 	    case jit_code_putargr_i:	case jit_code_putargr_ui:
-		jit_stxi_i(v->u.w, JIT_FP, u);			break;
+		node = jit_stxi_i(v->u.w, JIT_FP, u);		break;
 	    case jit_code_putargr_l:
-		jit_stxi(v->u.w, JIT_FP, u);			break;
+		node = jit_stxi(v->u.w, JIT_FP, u);		break;
 #  endif
 	    default:			abort();		break;
 	}
 #else
-	jit_stxi(v->u.w, JIT_FP, u);
+	node = jit_stxi(v->u.w, JIT_FP, u);
 #endif
+	jit_link_alist(node);
     }
     jit_dec_synth();
 }
@@ -659,28 +694,30 @@ _jit_putargi(jit_state_t *_jit, jit_word_t u, jit_node_t *v, jit_code_t code)
     if (jit_arg_reg_p(v->u.w))
 	jit_movi(JIT_RA0 - v->u.w, u);
     else {
+	jit_node_t	*node;
 	regno = jit_get_reg(jit_class_gpr);
 	jit_movi(regno, u);
 #if PACKED_STACK
 	switch (code) {
 	    case jit_code_putargi_c:	case jit_code_putargi_uc:
-		jit_stxi_c(v->u.w, JIT_FP, regno);		break;
+		node = jit_stxi_c(v->u.w, JIT_FP, regno);	break;
 	    case jit_code_putargi_s:	case jit_code_putargi_us:
-		jit_stxi_s(v->u.w, JIT_FP, regno);		break;
+		node = jit_stxi_s(v->u.w, JIT_FP, regno);	break;
 #  if __WORDSIZE == 32
 	    case jit_code_putargi_i:
-		jit_stxi(v->u.w, JIT_FP, regno);		break;
+		node = jit_stxi(v->u.w, JIT_FP, regno);		break;
 #  else
 	    case jit_code_putargi_i:	case jit_code_putargi_ui:
-		jit_stxi_i(v->u.w, JIT_FP, regno);		break;
+		node = jit_stxi_i(v->u.w, JIT_FP, regno);	break;
 	    case jit_code_putargi_l:
-		jit_stxi(v->u.w, JIT_FP, regno);		break;
+		node = jit_stxi(v->u.w, JIT_FP, regno);		break;
 #  endif
 	    default:			abort();		break;
 	}
 #else
-	jit_stxi(v->u.w, JIT_FP, regno);
+	node = jit_stxi(v->u.w, JIT_FP, regno);
 #endif
+	jit_link_alist(node);
 	jit_unget_reg(regno);
     }
     jit_dec_synth();
@@ -693,8 +730,10 @@ _jit_getarg_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     jit_inc_synth_wp(getarg_f, u, v);
     if (jit_arg_reg_p(v->u.w))
 	jit_movr_f(u, JIT_FA0 - v->u.w);
-    else
-	jit_ldxi_f(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_f(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -705,8 +744,10 @@ _jit_putargr_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     jit_inc_synth_wp(putargr_f, u, v);
     if (jit_arg_f_reg_p(v->u.w))
 	jit_movr_f(JIT_FA0 - v->u.w, u);
-    else
-	jit_stxi_f(v->u.w, JIT_FP, u);
+    else {
+	jit_node_t	*node = jit_stxi_f(v->u.w, JIT_FP, u);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -719,9 +760,11 @@ _jit_putargi_f(jit_state_t *_jit, jit_float32_t u, jit_node_t *v)
     if (jit_arg_f_reg_p(v->u.w))
 	jit_movi_f(JIT_FA0 - v->u.w, u);
     else {
+	jit_node_t	*node;
 	regno = jit_get_reg(jit_class_fpr);
 	jit_movi_f(regno, u);
-	jit_stxi_f(v->u.w, JIT_FP, regno);
+	node = jit_stxi_f(v->u.w, JIT_FP, regno);
+	jit_link_alist(node);
 	jit_unget_reg(regno);
     }
     jit_dec_synth();
@@ -734,8 +777,10 @@ _jit_getarg_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     jit_inc_synth_wp(getarg_d, u, v);
     if (jit_arg_f_reg_p(v->u.w))
 	jit_movr_d(u, JIT_FA0 - v->u.w);
-    else
-	jit_ldxi_d(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_d(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -746,8 +791,10 @@ _jit_putargr_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     jit_inc_synth_wp(putargr_d, u, v);
     if (jit_arg_reg_p(v->u.w))
 	jit_movr_d(JIT_FA0 - v->u.w, u);
-    else
-	jit_stxi_d(v->u.w, JIT_FP, u);
+    else {
+	jit_node_t	*node = jit_stxi_d(v->u.w, JIT_FP, u);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -760,9 +807,11 @@ _jit_putargi_d(jit_state_t *_jit, jit_float64_t u, jit_node_t *v)
     if (jit_arg_reg_p(v->u.w))
 	jit_movi_d(JIT_FA0 - v->u.w, u);
     else {
+	jit_node_t	*node;
 	regno = jit_get_reg(jit_class_fpr);
 	jit_movi_d(regno, u);
-	jit_stxi_d(v->u.w, JIT_FP, regno);
+	node = jit_stxi_d(v->u.w, JIT_FP, regno);
+	jit_link_alist(node);
 	jit_unget_reg(regno);
     }
     jit_dec_synth();
@@ -836,6 +885,7 @@ _jit_pushargr(jit_state_t *_jit, jit_int32_t u, jit_code_t code)
 	jit_stxi(_jitc->function->call.size, JIT_SP, u);
 	_jitc->function->call.size += sizeof(jit_word_t);
 #endif
+	jit_check_frame();
     }
     jit_dec_synth();
 }
@@ -911,6 +961,7 @@ _jit_pushargi(jit_state_t *_jit, jit_word_t u, jit_code_t code)
 	_jitc->function->call.size += sizeof(jit_word_t);
 #endif
 	jit_unget_reg(regno);
+	jit_check_frame();
     }
     jit_dec_synth();
 }
@@ -943,6 +994,7 @@ _jit_pushargr_f(jit_state_t *_jit, jit_int32_t u)
 	jit_stxi_f(_jitc->function->call.size, JIT_SP, u);
 	_jitc->function->call.size += sizeof(jit_word_t);
 #endif
+	jit_check_frame();
     }
     jit_dec_synth();
 }
@@ -979,6 +1031,7 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
 	_jitc->function->call.size += sizeof(jit_word_t);
 #endif
 	jit_unget_reg(regno);
+	jit_check_frame();
     }
     jit_dec_synth();
 }
@@ -1008,6 +1061,7 @@ _jit_pushargr_d(jit_state_t *_jit, jit_int32_t u)
 #endif
 	jit_stxi_d(_jitc->function->call.size, JIT_SP, u);
 	_jitc->function->call.size += sizeof(jit_float64_t);
+	jit_check_frame();
     }
     jit_dec_synth();
 }
@@ -1041,6 +1095,7 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
  	jit_stxi_d(_jitc->function->call.size, JIT_SP, regno);
 	jit_unget_reg(regno);
 	_jitc->function->call.size += sizeof(jit_float64_t);
+	jit_check_frame();
     }
     jit_dec_synth();
 }
@@ -1069,6 +1124,7 @@ _jit_finishr(jit_state_t *_jit, jit_int32_t r0)
 {
     jit_node_t		*node;
     assert(_jitc->function);
+    jit_check_frame();
     jit_inc_synth_w(finishr, r0);
 #if PACKED_STACK
     _jitc->function->call.size +=
@@ -1090,6 +1146,7 @@ _jit_finishi(jit_state_t *_jit, jit_pointer_t i0)
 {
     jit_node_t		*node;
     assert(_jitc->function);
+    jit_check_frame();
     jit_inc_synth_w(finishi, (jit_word_t)i0);
 #if PACKED_STACK
     _jitc->function->call.size +=
@@ -1387,6 +1444,10 @@ _emit_code(jit_state_t *_jit)
 		case_rrw(rsh, _u);
 		case_rr(neg,);
 		case_rr(com,);
+		case_rr(clo,);
+		case_rr(clz,);
+		case_rr(cto,);
+		case_rr(ctz,);
 		case_rrr(and,);
 		case_rrw(and,);
 		case_rrr(or,);
@@ -1708,6 +1769,7 @@ _emit_code(jit_state_t *_jit)
 		case_brr(bunord, _d);
 		case_brd(bunord);
 	    case jit_code_jmpr:
+		jit_check_frame();
 		jmpr(rn(node->u.w));
 		break;
 	    case jit_code_jmpi:
@@ -1727,13 +1789,17 @@ _emit_code(jit_state_t *_jit)
 			patch(word, node);
 		    }
 		}
-		else
+		else {
+		    jit_check_frame();
 		    jmpi(node->u.w);
+		}
 		break;
 	    case jit_code_callr:
+		jit_check_frame();
 		callr(rn(node->u.w));
 		break;
 	    case jit_code_calli:
+		jit_check_frame();
 		if (node->flag & jit_flag_node) {
 		    temp = node->u.n;
 		    assert(temp->code == jit_code_label ||
@@ -1763,6 +1829,8 @@ _emit_code(jit_state_t *_jit)
 #endif
 		undo.patch_offset = _jitc->patches.offset;
 	    restart_function:
+		compute_framesize();
+		patch_alist(0);
 		_jitc->again = 0;
 		prolog(node);
 		break;
@@ -1783,12 +1851,17 @@ _emit_code(jit_state_t *_jit)
 		     * the reason of the undo. */
 		    undo.func.self.aoff = _jitc->function->frame +
 			_jitc->function->self.aoff;
+		    undo.func.need_frame = _jitc->function->need_frame;
 		    jit_regset_set(&undo.func.regset, &_jitc->function->regset);
+		    /* allocar information also does not need to be undone */
+		    undo.func.aoffoff = _jitc->function->aoffoff;
+		    undo.func.allocar = _jitc->function->allocar;
 		    memcpy(_jitc->function, &undo.func, sizeof(undo.func));
 #if DEVEL_DISASSEMBLER
 		    prevw = undo.prevw;
 #endif
 		    _jitc->patches.offset = undo.patch_offset;
+		    patch_alist(1);
 		    goto restart_function;
 		}
 		/* remember label is defined */
@@ -1936,6 +2009,23 @@ _emit_stxi_d(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_int32_t r1)
     stxi_d(i0, rn(r0), rn(r1));
 }
 
+static void
+_compute_framesize(jit_state_t *_jit)
+{
+    jit_int32_t		reg;
+    _jitc->framesize = 16;	/* ra+fp */
+    for (reg = 0; reg < jit_size(iregs); reg++)
+	if (jit_regset_tstbit(&_jitc->function->regset, iregs[reg]))
+	    _jitc->framesize += sizeof(jit_word_t);
+
+    for (reg = 0; reg < jit_size(fregs); reg++)
+	if (jit_regset_tstbit(&_jitc->function->regset, fregs[reg]))
+	    _jitc->framesize += sizeof(jit_float64_t);
+
+    /* Make sure functions called have a 16 byte aligned stack */
+    _jitc->framesize = (_jitc->framesize + 15) & -16;
+}
+
 static void
 _patch(jit_state_t *_jit, jit_word_t instr, jit_node_t *node)
 {
